<?php

use Drupal\Signify\ChecksumList;
use Drupal\Signify\Verifier;
use Drush\Log\LogLevel;
use phpseclib3\Net\SFTP;

/**
 * Implementation of hook_drush_help().
 */
function drupalorg_drush_help($section) {
  switch ($section) {
    case 'drupalorg:rebuild-stats':
      return dt('Update home page stats.');
  }
}

/**
 * Implementation of hook_drush_command();
 */
function drupalorg_drush_command() {
  return [
    'rebuild-stats' => [
      'description' => 'Update home page stats.',
    ],
    'drupalorg-org-credit' => [
      'description' => 'Update organization credit totals.',
    ],
    'drupalorg-integrity' => [
      'description' => 'Integrity check for a non-ideal world.',
    ],
    'drupalorg-release-hashes' => [
      'description' => 'Generate release contents hashes for a single release.',
      'arguments' => [
        'project' => 'Project, either a Node ID or machine name.',
        'version' => 'Version string of release.',
      ],
    ],
    'drupalorg-resign-files' => [
      'description' => 'Re-sign csig files with a new intermediate key. File list is null-separated from stdin.'
    ],
    'drupalorg-release-send-rugged' => [
      'description' => 'Send a packaged release to rugged for TUF.',
      'arguments' => [
        'project' => 'Project, either a Node ID or machine name.',
      ],
      'options' => [
        'version' => 'Only sign a specific version.',
        'skip-packages-json' => 'Skip sending packages.json, for backfilling data.',
      ],
    ],
    'drupalorg-update-announcements-json' => [
      'description' => 'Publish and updates to the announcements.json API.',
      'options' => [
        'test' => 'Write to the announcements-this-is-only-a-test.json endpoint.',
      ],
    ],
    'drupalorg-update-psa-json' => [
      'description' => 'Publish an update to the psa.json API.',
      'arguments' => [
        'nids' => 'Comma-separated PSA node IDs to include.',
      ],
      'options' => [
        'test' => 'Write to the psa-this-is-only-a-test.json endpoint.',
      ],
    ],
    'drupalorg-create-quasi-patch' => [
      'description' => 'Generate a quasi-patch for a single release. ⚠️ Should only be called for combinations in the drupalorg_in_place_updates table; usually triggered via queue running.',
      'arguments' => [
        'from_release_nid' => 'From release node ID.',
        'to_release_nid' => 'To release node ID.',
      ],
    ],
    'drupalorg-packaging-status' => [
      'description' => 'Packaging pipeline should call this step upon completion to notify drupal.org that the job has finished running.',
      'arguments' => [
        'job_id' => 'drupalorg_packaging_job ID',
        'status' => 'New status of job',
        'message' => 'Relevant success/error message',
      ],
    ],
    'drupalorg-packaging-resend' => [
      'description' => 'Re-sends queued jobs to jenkins',
    ],
    'drupalorg-repath-docs' => [
      'description' => 'Re-pathauto a tree of documentation',
      'arguments' => [
        'nid' => 'Documentation guide node ID',
      ],
    ],
    'drupalorg-tugboat-register-project' => [
      'description' => 'Registers a project at Tugboat',
      'arguments' => [
        'project_namespace' => 'The project namespace, like project.',
        'project_name' => 'The project name, like drupal.',
      ],
    ],
    'drupalorg-tugboat-build-base-preview' => [
      'description' => 'Builds a base preview for a project and branch',
      'arguments' => [
        'project_name' => 'The project name, like drupal for Drupal core. This is used to pick a configuration from https://github.com/TugboatQA/drupalorg.',
        'branch' => 'The branch to build as a base preview',
      ],
    ],
    'drupalorg-purge-fastly' => [
      'description' => 'Purges urls in fastly',
      'arguments' => [
        'urls' => 'The urls you want purged (space delimeted if more than one).',
      ],
    ],
    'drupalorg-project-packagist-submitter' => [
      'description' => 'Decides if a project should get releses submitted to packagist',
      'arguments' => [
        'release' => 'The release nid',
      ],
    ],
    'drupalorg-mark-revoked-project-releases-insecure' => [
      'description' => 'Marks releases for projects with revoked security advisory coverage as insecure',
    ],
    'drupalorg_uid_salesforce_push' => [
      'description' => 'If an account has an active membership, push it\'s uid up to salesforce',
    ],
    'drupalorg-update-project-analysis-data' => [
      'description' => 'Updates files/project_analysis/allprojects.tsv for https://www.drupal.org/project/project_analysis',
    ],
    'drupalorg-unsupport-stale-core-compatibility' => [
      'description' => 'Marks contrib releases with stale core compatibility as unsupported.',
      'arguments' => [
        'core_constraint' => 'Version constriant for Drupal core that is supported, such as "^10"',
      ],
      'options' => [
        'maintainers-for-email' => 'Get maintainers of branches that will be unsupported',
        'no-dry-run' => 'Actually mark releases as unsupported',
      ],
    ],
    'drupalorg-migrate-org-credits' => [
      'description' => 'Moves credits from an org to another org',
      'arguments' => [
        'origin_organization_nid' => 'Origin organization’s node id',
        'target_organization_nid' => 'Target organizations’s node id',
        'date' => 'Y-M-D of merger'
      ],
    ],
    'drupalorg-sync-keycloak-user-delete' => [
      'description' => 'Synchronize user deletes from drupal to keycloak',
    ],
    'drupalorg-keycloak-user-by-uuid' => [
      'description' => 'Get keycloak user info by its UUID',
      'arguments' => [
        'uuid' => 'The keycloak uuid',
      ],
    ],
    'drupalorg-keycloak-admin-users-integrity' => [
      'description' => 'Integrity check migration of admin users',
    ],
    'drupalorg-keycloak-random-users-integrity' => [
      'description' => 'Integrity check migration of random users',
    ],
    'drupalorg-keycloak-users-count' => [
      'description' => 'Reports total users in drupal and keycloak',
    ],
  ];
}

function drush_drupalorg_rebuild_stats() {
  drupalorg_get_activity(TRUE);
}

/**
 * @param $account
 * @param $sf_object
 * @param $sf_mapping
 *
 * @return void
 * @throws \EntityMetadataWrapperException
 */
function drush_drupalorg_uid_salesforce_push() {
  // Query for users with a membership status
  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'user')
    ->propertyCondition('status', 1)
    ->fieldCondition('field_da_ind_membership', 'value', NULL, 'IS NOT NULL')
    ->fieldCondition('field_da_ind_membership', 'value', '', '!=');
  $result = $query->execute();
  foreach ($result['user'] as $uid => $value) {
    $user_list[$value->uid] = $value->uid;
  }
  $users = user_load_multiple($user_list);
  foreach ($users as $account) {
    $wrapper = entity_metadata_wrapper('user', $account);

    if ($wrapper->field_last_name->value() != '') {
      $entity_type = 'user';
      $entity = $account;
      $entity->force_sf_push = 1;
      $sf_sync_trigger = SALESFORCE_MAPPING_SYNC_DRUPAL_UPDATE;
      $mapping = salesforce_mapping_load('uidpush');
      $sf_push = salesforce_push_sync_rest($entity_type, $entity, $mapping, $sf_sync_trigger);
      if ($sf_push) {
        watchdog('drupalorg salesforce push', 'Salesforce push successful for @name', ['@name' => $account->name], WATCHDOG_INFO);
      }
    }
  }
}

/**
 * @param $origin_organization_nid
 * @param $target_organization_nid
 * @return void
 */
function drush_drupalorg_migrate_org_credits($origin_organization_nid, $target_organization_nid, $date) {
  $orgs = node_load_multiple([$origin_organization_nid, $target_organization_nid]);

  if (count($orgs) !== 2 || $orgs[$origin_organization_nid]->type !== 'organization' || $orgs[$target_organization_nid]->type !== 'organization') {
    drush_log(dt('Please double check your organization nids.'), LogLevel::ERROR);
    return;
  }

  $effective_date = (new DateTime($date))->sub(new DateInterval('P90D'))->format('U');
  drush_log(dt('Migrating issue credits from @origin to @target for issues after @effective_date including open issues', [
    '@origin' => $orgs[$origin_organization_nid]->title,
    '@target' => $orgs[$target_organization_nid]->title,
    '@effective_date' => format_date($effective_date),
  ]));

  $cids = db_query('SELECT max(c.cid) FROM {comment} c INNER JOIN {users} u ON u.uid = c.uid INNER JOIN {field_data_field_organizations} fdf_o ON fdf_o.entity_id = u.uid INNER JOIN {field_data_field_current} fdf_c ON fdf_c.entity_id = fdf_o.field_organizations_value AND fdf_c.field_current_value = 1 INNER JOIN {field_data_field_organization_reference} fdf_or ON fdf_or.entity_id = fdf_o.field_organizations_value AND fdf_or.field_organization_reference_target_id = :origin_organization_nid INNER JOIN {field_data_field_issue_status} fdf_is ON fdf_is.entity_id = c.nid INNER JOIN {field_data_field_attribute_contribution_to} fdf_act ON fdf_act.entity_id = c.cid AND fdf_act.field_attribute_contribution_to_target_id = :origin_organization_nid INNER JOIN {field_data_field_issue_last_status_change} fdf_ilsc ON fdf_ilsc.entity_id = c.nid WHERE fdf_ilsc.field_issue_last_status_change_value > :effective_date OR fdf_is.field_issue_status_value IN (:open_states) GROUP BY c.nid, c.uid', [
    ':origin_organization_nid' => $origin_organization_nid,
    ':effective_date' => $effective_date,
    ':open_states' => project_issue_open_states(),
  ])->fetchCol();

  if (drush_confirm(dt('@count comments will be updated, would you like to proceed?', ['@count' => count($cids)]))) {
    foreach (entity_load('comment', $cids) as $comment) {
      foreach ($comment->field_attribute_contribution_to[LANGUAGE_NONE] as $key => $value) {
        if ($comment->field_attribute_contribution_to[LANGUAGE_NONE][$key]['target_id'] == $origin_organization_nid) {
          drush_log(dt('Updating @url', ['@url' => url('comment/' . $comment->cid, ['absolute' => TRUE])]));
          $comment->field_attribute_contribution_to[LANGUAGE_NONE][$key]['target_id'] = $target_organization_nid;
          comment_save($comment);
          continue;
        }
      }
    }
  }
}

/**
 * Command callback. Count issue credits and store count on organization profile.
 */
function drush_drupalorg_org_credit() {
  $weights = variable_get('drupalorg_org_credit_weights');
  $credits = [];
  $usage_weights = [];
  $contrib_roles_weighted_score = 0;
  //field_drupalorg_rank_components list
  //0: weighted issue credits
  //1: association partnerships
  //2: association memberships
  //3: individual memberships
  //4: case studies
  //5: project(s) supported
  //6: contributor roles
  //7: contribution without issue credits
  //8: contribution with 3 months of issue credit
  //9: contribution with 12 months of issue credit
  //10: Event contribution

  // Get all users with VCS edit access
  $maintainer_list = db_query("SELECT DISTINCT users.uid FROM {versioncontrol_auth_account} LEFT JOIN {users} ON versioncontrol_auth_account.uid = users.uid")->fetchCol();

  // Load up roles to create an array of nids and weights
  $role_nids_for_exclusion = [
    3219628, // Project contributor
    3170722, // Project moderator
  ];
  $query = new EntityFieldQuery();
  $result = $query->entityCondition('entity_type', 'node')
    ->entityCondition('bundle', 'contributor_role')
    ->propertyCondition('nid', $role_nids_for_exclusion, 'NOT IN')
    ->execute();
  $nids = array_keys($result['node']);
  if (!empty($result['node'])) {
    $roles = node_load_multiple($nids);

    foreach ($roles as $role) {
      // Some roles have multiple contribution areas
      $contribution_area_sum = 0;
      if (!empty($role->field_contribution_areas[LANGUAGE_NONE])) {
        foreach ($role->field_contribution_areas[LANGUAGE_NONE] as $contribution_area) {
          $contribution_area_sum += $weights['contribution role']['contrib area'][taxonomy_term_load($contribution_area['tid'])->name];
        }
      }

      $role_base = $weights['contribution role']['base'];
      if (isset($role->field_duration[LANGUAGE_NONE][0]['value'])) {
        $role_duration = $weights['contribution role']['time duration'][$role->field_duration[LANGUAGE_NONE][0]['value']];
      }
      else {
        $role_duration = 0;
      }
      if (isset($role->field_average_time[LANGUAGE_NONE][0]['value'])) {
        $role_monthly = $weights['contribution role']['time monthly'][$role->field_average_time[LANGUAGE_NONE][0]['value']];
      }
      else {
        $role_monthly = 0;
      }

      $role_math = $role_base + $role_duration + $role_monthly + $contribution_area_sum;
      $roles_weighted[$role->nid] = array(
        'nid' => $role->nid,
        'name' => $role->title,
        'role weight' => $role_math,
      );
    }
  }

  // Check events for sponsorship and grant credit if they are within the past
  // year.
  $one_year_ago = (new DateTimeImmutable('-1 year'))->format('c');
  $query = new EntityFieldQuery();
  $result = $query->entityCondition('entity_type', 'node')
    ->propertyCondition('type', 'event')
    ->fieldCondition('field_date_of_event', 'value', $one_year_ago, '>=')
    ->execute();
  foreach (node_load_multiple(array_keys($result['node'])) as $event) {
    if (!empty($event->field_event_sponsors[LANGUAGE_NONE])) {
      foreach($event->field_event_sponsors[LANGUAGE_NONE] as $key => $sponsor) {
        $events[$sponsor['target_id']][$event->field_event_type[LANGUAGE_NONE][0]['value']][] = $event->nid;
      }
    }
  }

  // Organization attribution.
  $result = db_query("SELECT fdfact.field_attribute_contribution_to_target_id AS organization, n.nid, fdf_p.field_project_target_id AS project, fdf_ilsc.field_issue_last_status_change_value > UNIX_TIMESTAMP(NOW() - INTERVAL 3 MONTH) AS recent FROM {node} n INNER JOIN {field_data_field_issue_credit} fdfic ON fdfic.entity_id = n.nid INNER JOIN {comment} c ON c.cid = fdfic.field_issue_credit_target_id AND c.status = 1 INNER JOIN {field_data_field_issue_status} fdfis ON fdfis.entity_id = n.nid AND fdfis.field_issue_status_value IN (2,7) INNER JOIN {field_data_field_attribute_as_volunteer} fdf_av ON fdf_av.entity_id = c.cid AND fdf_av.field_attribute_as_volunteer_value = 0 INNER JOIN {field_data_field_attribute_contribution_to} fdfact ON fdfact.entity_id = c.cid INNER JOIN {field_data_field_project} fdf_p ON fdf_p.entity_id = n.nid LEFT JOIN {field_data_field_project_type} fdf_pt ON fdf_pt.entity_id = fdf_p.field_project_target_id INNER JOIN {field_data_field_issue_last_status_change} fdf_ilsc ON fdf_ilsc.entity_id = n.nid WHERE n.type IN (:project_issue_types) AND n.status = 1 AND (fdf_pt.field_project_type_value = 'full' OR fdf_pt.field_project_type_value IS NULL) AND fdf_ilsc.field_issue_last_status_change_value > UNIX_TIMESTAMP(NOW() - INTERVAL 1 YEAR)", [':project_issue_types' => project_issue_issue_node_types()]);
  foreach ($result as $row) {
    if ($row->recent){
      $credits[$row->organization][$row->nid] = $row->project;
    }
    $year_credits[$row->organization][$row->nid] = $row->project;
  }

  // Customer attribution.
  $result = db_query("SELECT fdf_fc.field_for_customer_target_id AS organization, n.nid, fdf_p.field_project_target_id AS project, fdf_ilsc.field_issue_last_status_change_value > UNIX_TIMESTAMP(NOW() - INTERVAL 3 MONTH) AS recent FROM {node} n INNER JOIN {field_data_field_issue_credit} fdfic ON fdfic.entity_id = n.nid INNER JOIN {comment} c ON c.cid = fdfic.field_issue_credit_target_id AND c.status = 1 INNER JOIN {field_data_field_issue_status} fdfis ON fdfis.entity_id = n.nid AND fdfis.field_issue_status_value IN (2,7) INNER JOIN {field_data_field_attribute_as_volunteer} fdf_av ON fdf_av.entity_id = c.cid AND fdf_av.field_attribute_as_volunteer_value = 0 INNER JOIN {field_data_field_for_customer} fdf_fc ON fdf_fc.entity_id = c.cid INNER JOIN {field_data_field_project} fdf_p ON fdf_p.entity_id = n.nid LEFT JOIN {field_data_field_project_type} fdf_pt ON fdf_pt.entity_id = fdf_p.field_project_target_id INNER JOIN {field_data_field_issue_last_status_change} fdf_ilsc ON fdf_ilsc.entity_id = n.nid WHERE n.type IN (:project_issue_types) AND n.status = 1 AND (fdf_pt.field_project_type_value = 'full' OR fdf_pt.field_project_type_value IS NULL) AND fdf_ilsc.field_issue_last_status_change_value > UNIX_TIMESTAMP(NOW() - INTERVAL 1 YEAR)", [':project_issue_types' => project_issue_issue_node_types()]);
  // De-dupe by issue nid.
  foreach ($result as $row) {
    if ($row->recent) {
      $credits[$row->organization][$row->nid] = $row->project;
    }
    $year_credits[$row->organization][$row->nid] = $row->project;
  }

  $maxval = array_fill(0, 8, 0) + ['total' => 0];
  // Update all organizations.
  $query = new EntityFieldQuery();
  $result = $query->entityCondition('entity_type', 'node')
    ->propertyCondition('type', 'organization')
    ->execute();
  $org_nodes = node_load_multiple(array_keys($result['node']));
  foreach ($org_nodes as $keyed) {
    $nodes[$keyed->title] = $keyed;
  }
  ksort($nodes);
  foreach ($nodes as $node) {
    $contribution_rank = 0;
    $wrapper = entity_metadata_wrapper('node', $node);
    $updated = FALSE;

    $issue_credits = isset($credits[$node->nid]) ? count($credits[$node->nid]) : 0;
    if ($wrapper->field_org_issue_credit_count->value() != $issue_credits) {
      $wrapper->field_org_issue_credit_count = $issue_credits;
      $updated = TRUE;
    }

    if (array_key_exists($node->nid, $year_credits) && ($wrapper->field_org_issue_credit_year->value() != count($year_credits[$node->nid]))) {
      drush_log('Updating year credit: ' . $node->title . ' from ' . $wrapper->field_org_issue_credit_year->value() . ' to ' . count($year_credits[$node->nid]));
      $wrapper->field_org_issue_credit_year = count($year_credits[$node->nid]);
      $updated = TRUE;
    }

    // Add issue credit, weighted by project usage for 12 months worth of credits.
    $year_issue_credit_weighted = 0;
    if (isset($year_credits[$node->nid])) {
      foreach ($year_credits[$node->nid] as $project_nid) {
        if (!isset($usage_weights[$project_nid])) {
          $usage_weights[$project_nid] = _drupalorg_weight_project($project_nid);
        }
        $year_issue_credit_weighted += $usage_weights[$project_nid];
      }
      $updated = TRUE;
    }
    $wrapper->field_weighted_org_issue_cred_yr = $weights['issue credit'] * $year_issue_credit_weighted;

    // Add issue credit, weighted by project usage.
    $issue_credit_weighted = 0;
    if (isset($credits[$node->nid])) {
      foreach ($credits[$node->nid] as $project_nid) {
        if (!isset($usage_weights[$project_nid])) {
          $usage_weights[$project_nid] = _drupalorg_weight_project($project_nid);
        }
        $issue_credit_weighted += $usage_weights[$project_nid];
      }
    }
    $wrapper->field_drupalorg_rank_components[0] = $weights['issue credit'] * $issue_credit_weighted;

    // Add Association partnerships.
    $wrapper->field_drupalorg_rank_components[1] = array_sum(array_intersect_key($weights['supporting partner'], array_flip($wrapper->field_organization_support->value())));
    $contribution_rank += $wrapper->field_drupalorg_rank_components[1]->value();

    // Add Association membership.
    $wrapper->field_drupalorg_rank_components[2] = $weights['membership'] * (bool) $wrapper->field_org_membership_status->value();
    $contribution_rank += $wrapper->field_drupalorg_rank_components[2]->value();

    // Add Association individual membership of people.
    $result = (new EntityFieldQuery())->entityCondition('entity_type', 'field_collection_item')
      ->fieldCondition('field_organization_reference', 'target_id', $node->nid)
      ->fieldCondition('field_current', 'value', 1)->execute();
    if (!empty($result['field_collection_item'])) {
      $wrapper->field_drupalorg_rank_components[3] = $weights['membership_individual'] * (new EntityFieldQuery())->entityCondition('entity_type', 'user')
          ->propertyCondition('status', 1)
          ->fieldCondition('field_da_ind_membership', 'value', NULL, 'IS NOT NULL')
          ->fieldCondition('field_da_ind_membership', 'value', '', '!=')
          ->fieldCondition('field_organizations', 'value', array_keys($result['field_collection_item']))
          ->count()->execute();
      $contribution_rank += $wrapper->field_drupalorg_rank_components[3]->value();
    }

    // Case Studies
    // Add D8 case studies.
    $case_study_rank = $weights['d8 case study'] * (new EntityFieldQuery())->entityCondition('entity_type', 'node')
        ->propertyCondition('type', 'casestudy')
        ->fieldCondition('field_case_organizations', 'target_id', $node->nid)
        ->fieldCondition('taxonomy_vocabulary_5', 'tid', 20236) // D8
        ->count()->execute();
    // Drupal 9 case studies
    $case_study_rank += $weights['d9 case study'] * (new EntityFieldQuery())->entityCondition('entity_type', 'node')
        ->propertyCondition('type', 'casestudy')
        ->fieldCondition('field_case_organizations', 'target_id', $node->nid)
        ->fieldCondition('taxonomy_vocabulary_5', 'tid', 194424) // D9
        ->count()->execute();
    // Drupal 10 case studies
    $case_study_rank += $weights['d10 case study'] * (new EntityFieldQuery())->entityCondition('entity_type', 'node')
        ->propertyCondition('type', 'casestudy')
        ->fieldCondition('field_case_organizations', 'target_id', $node->nid)
        ->fieldCondition('taxonomy_vocabulary_5', 'tid', 195821) // D10
        ->count()->execute();
    $wrapper->field_drupalorg_rank_components[4] = $case_study_rank;
    $contribution_rank += $wrapper->field_drupalorg_rank_components[4]->value();

    // Add Projects supported.
    $query = new EntityFieldQuery();
    $wrapper->field_drupalorg_rank_components[5] = $weights['project supported'] * $query->entityCondition('entity_type', 'field_collection_item')
        ->fieldCondition('field_supporting_organization', 'target_id', $node->nid)
        ->propertyCondition('archived', FALSE)
        ->count()->execute();
    $contribution_rank += $wrapper->field_drupalorg_rank_components[5]->value();

    // Add contribution roles
    $query = new EntityFieldQuery();
    $query->entityCondition('entity_type', 'field_collection_item')
      ->entityCondition('bundle', 'field_contribution_role')
      ->fieldCondition('field_organization_reference', 'target_id', $node->nid)
      ->propertyCondition('archived', FALSE);

    $org_supported_roles = $query->execute();
    $contrib_roles_weighted_score = 0;
    // If we have roles, lets get their ID's, so we can assign weighted values
    if (isset($org_supported_roles['field_collection_item'])) {
      foreach (field_collection_item_load_multiple(array_keys($org_supported_roles['field_collection_item'])) as $collection) {
        $collection_wrapper = entity_metadata_wrapper('field_collection_item', $collection);
        if ($collection_role = $collection_wrapper->field_con_role->value()) {
          // Exclude “Contributed module, theme, or distribution maintainer”
          // contributor role if the user is not actually a maintainer of
          // anything. And exclude any without weights assigned.
          if (($collection_role->nid == 3218720 && !in_array($collection->hostEntityId(), $maintainer_list)) || !isset($roles_weighted[$collection_role->nid])) {
            continue;
          }

          // If the role is not current, it earns half credit.
          if (($collection_wrapper->field_role_end_date->value() != NULL) && ($collection_wrapper->field_role_end_date->value() <= REQUEST_TIME)) {
            $contrib_roles_weighted_score += $roles_weighted[$collection_role->nid]['role weight'] * $weights['contribution role']['time']['expired'];
          }
          elseif ($collection_wrapper->field_role_start_date == NULL) {
            $contrib_roles_weighted_score += $roles_weighted[$collection_role->nid]['role weight'] * $weights['contribution role']['time']['expired'];
          }
          else {
            $contrib_roles_weighted_score += $roles_weighted[$collection_role->nid]['role weight'];
          }
        }
      }
    }
    if ($wrapper->field_drupalorg_rank_components[6]->value() != $contrib_roles_weighted_score) {
      $wrapper->field_drupalorg_rank_components[6] = $contrib_roles_weighted_score;
      $updated = TRUE;
    }
    $contribution_rank += $wrapper->field_drupalorg_rank_components[6]->value();

    // Include special programs field
    if (!empty($wrapper->field_org_special_program->value())) {
      $contribution_rank += $wrapper->field_org_special_program->value();
      $updated = TRUE;
    }

    // Include events sponsorship.
    $event_weights = variable_get('drupalorg_events_sponsorship');
    $weighted_event_credits = 0;
    if (!empty($events[$node->nid])) {
      foreach ($events[$node->nid] as $event_type => $event_nids) {
        $weighted_event_credits += count($event_nids) * $event_weights[$event_type];
      }
    }
    if ($wrapper->field_drupalorg_rank_components[10]->value() != $weighted_event_credits) {
      $wrapper->field_drupalorg_rank_components[10] = $weighted_event_credits;
      $updated = TRUE;
    }
    $contribution_rank += $weighted_event_credits;

    // Contribution without issue credits
    if (empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][7]['value']) || ($node->field_drupalorg_rank_components[LANGUAGE_NONE][7]['value'] != $contribution_rank)) {
      $wrapper->field_drupalorg_rank_components[7] = $contribution_rank;
      $updated = TRUE;
    }

    // Contribution with 3 months of issue credits
    $sum = $contribution_rank + $node->field_drupalorg_rank_components[LANGUAGE_NONE][0]['value'];
    if (empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][8]['value']) || ($node->field_drupalorg_rank_components[LANGUAGE_NONE][8]['value'] != $sum)) {
      $wrapper->field_drupalorg_rank_components[8] = $sum;
      $updated = TRUE;
    }

    // Contribution with 12 months of issue credits
    $sum = $contribution_rank + $node->field_weighted_org_issue_cred_yr[LANGUAGE_NONE][0]['value'];
    if(empty($node->field_drupalorg_rank_components[LANGUAGE_NONE][9]['value']) || ($node->field_drupalorg_rank_components[LANGUAGE_NONE][9]['value'] != $sum)) {
      $wrapper->field_drupalorg_rank_components[9] = $sum;
      $updated = TRUE;
    }

    // Round & store 12 month value for the marketplace.
    $new_rank = round($wrapper->field_drupalorg_rank_components[9]->value() ?: 0);
    if ($wrapper->field_org_contribution_rank->value() != $new_rank) {
      $wrapper->field_org_contribution_rank = $new_rank;
      $updated = TRUE;
    }

    $mkt_rank = $wrapper->field_org_marketplace_rank->value();
    $new_mkt_rank = [];
    if (!empty($wrapper->field_organization_support->value())) {
      foreach ($wrapper->field_organization_support->value() as $value) {
        if (array_key_exists($value, $weights['supporting partner market'])) {
          $new_mkt_rank[$value] = $weights['supporting partner market'][$value];
        }
      }
    }
    $new_mkt_rank_sum = array_sum($new_mkt_rank);

    // Marketplace sorting
    if ($new_mkt_rank_sum != $mkt_rank) {
      $wrapper->field_org_marketplace_rank = $new_mkt_rank_sum;
      $updated = TRUE;
    }

    if ($updated) {
      drush_log(dt('Updating @title', ['@title' => $node->title]));
      field_attach_presave('node', $node);
      field_attach_update('node', $node);
    }
    else {
      drush_log(dt('No update for @title', ['@title' => $node->title]));
    }
  }
}

/**
 * Run the formula to translate project usage into a weight multiplier after
 * checking for an override.
 */
function _drupalorg_weight_project($project_nid) {
  static $weights;
  if (!isset($weights)) {
    $weights = variable_get('drupalorg_org_credit_weights');
  }

  // Community projects
  $project = node_load($project_nid);

  // Project usage override
  if (isset($weights['issue credit project usage override'][$project->nid])) {
    return _drupalorg_weight_usage($weights['issue credit project usage override'][$project->nid]);
  }
  // Project type override
  elseif (isset($weights['issue credit project type override'][$project->type])) {
    return $weights['issue credit project type override'][$project->type];
  }
  else {
    $usage = project_usage_get_project_total_usage($project->nid);
    return _drupalorg_weight_usage($usage);
  }
}

/**
 * Run this to purge URLs from fastly
 */
function drush_drupalorg_purge_fastly($urls) {
  foreach (func_get_args() as $url) {
    drupalorg_crosssite_fastly_purge_url($url);
  }
}

/**
 * Checks if a package should be submitted to packagist
 */

function drush_drupalorg_project_packagist_submitter($release_nid) {
  $release_node = node_load($release_nid);
  $project_node = node_load(project_release_get_release_project_nid($release_node));
  $return = FALSE;
  //  1. This should only run on general_project project types.
  if ($project_node->type == 'project_general') {
    // Get data from repo
    drush_log(dt('Get composer.json from ') . $project_node->title);
    try {
      // TODO: Make work with dev.drupalcode
      $client = new GuzzleHttp\Client([
        'base_uri' => 'https://git.drupalcode.org',
        'timeout' => 30.0,
      ]);
      // Strip -dev out of versions because it breaks drupalcode.org URLs
      if (strpos($release_node->field_release_version[LANGUAGE_NONE][0]['value'], '-dev')) {
        $version = trim($release_node->field_release_version[LANGUAGE_NONE][0]['value'], '-dev');
      }
      else {
        $version = $release_node->field_release_version[LANGUAGE_NONE][0]['value'];
      }
      // Set request URL
      $composer_json_url = '/project/' . $project_node->field_project_machine_name[LANGUAGE_NONE][0]['value'] . '/-/raw/' . $version . '/composer.json?inline=false';
      // Make request and assign it to a variable
      $composer_json = json_decode($client->request('GET', $composer_json_url, [])
        ->getbody(), TRUE);

      //  3. It should only run if the composer.json in the root of the project is named in the drupal namespace and matches the project machine name
      if ($composer_json['name'] != 'drupal/' . $project_node->field_project_machine_name[LANGUAGE_NONE][0]['value']) {
        $return = 1;
        throw new Exception('Machine name of project does not match composer.json name');
      }
      //4. We should not submit something that collides with the drupal/* namespace on drupal.org
      $drupalorg_namespace = db_query('SELECT package_namespace FROM project_composer_namespace_map pcnm WHERE pcnm.package_namespace = :machine_name', [
        ':machine_name' => $project_node->field_project_machine_name[LANGUAGE_NONE][0]['value'],
      ])->rowCount();

      if ($drupalorg_namespace > 0) {
        throw new Exception('Project collides with the drupal/* namespace on drupal.org');
        $return = 1;
      }

      // We should not submit something on the reserved namespace list
      $reserved_namespaces = [
        "at_base",
        "coder",
        "commerce_taxcloud",
        "console",
        "console-ar",
        "console-ca",
        "console-core",
        "console-develop",
        "console-dotenv",
        "console-en",
        "console-es",
        "console-extend",
        "console-extend-example",
        "console-extend-plugin",
        "console-fa-ir",
        "console-fr",
        "console-he",
        "console-hi",
        "console-hu",
        "console-id",
        "console-ja",
        "console-ko",
        "console-launcher",
        "console-mr",
        "console-pa",
        "console-pt-br",
        "console-ro",
        "console-ru",
        "console-ur",
        "console-vn",
        "console-yaml",
        "console-zh-hans",
        "console-zh-hant",
        "cors",
        "currency",
        "d7-plugin",
        "default_content",
        "dev-dependencies",
        "drupal",
        "drupal-builder",
        "drupal-core-strict",
        "drupal-driver",
        "drupal-extension",
        "drupal-library-installer-plugin",
        "drupal-scaffold",
        "drupalctl",
        "dynamic_entity_reference",
        "entity_views",
        "legacy-project",
        "legacy-scaffold-assets",
        "marconi",
        "micro",
        "opcache",
        "openstack_queues",
        "panopoly-framework",
        "parse-composer",
        "payment",
        "php-signify",
        "pinned-dev-dependencies",
        "plugin_selector",
        "predis",
        "recommended-project",
        "root",
        "search_api_elasticsearch",
        "spaces_og_accelerated",
        "stringoverrides",
        "tangler",
        "tqextension",
      ];
      foreach ($reserved_namespaces as $reserved_namespace) {
        if ($reserved_namespace == $project_node->field_project_machine_name['und'][0]['value']) {
          throw new Exception('Project collides with a reserved namespace');
          $return = FALSE;
        }
      }
      // We should not submit a project where core is the beginning of the name
      $core_check = strpos($project_node->field_project_machine_name[LANGUAGE_NONE][0]['value'], 'core');
      if ($core_check === 0) {
        throw new Exception('Project collides with core');
        $return = FALSE;
      }

      $return = TRUE;
    } catch (Exception $exception) {
      drush_set_error('Exception', $exception->getMessage());
      $return = FALSE;
    }
  } else {
    drush_log(dt('Project is not a general_project: @project_name', ['@project_name' => $project_node->name]),'error');

    drush_set_error('Exception', 'Project is not a general_project');
  }
  return $return;
}
/**
 * Run the formula to translate project usage into a weight multiplier if no
 * override.
 */
function _drupalorg_weight_usage($usage) {
  static $weights;
  if (!isset($weights)) {
    $weights = variable_get('drupalorg_org_credit_weights');
  }

  // No issue credit/project override
  $issue_credit_weight = $usage / $weights['issue credit project divisor'];

  if ($issue_credit_weight < 1) {
    return $issue_credit_weight + 1;
  }
  else {
    return $issue_credit_weight + 2;
  }
}

/**
 * Report on queries that should not have results, but do.
 */
function drush_drupalorg_integrity() {
  // Queries that should not have results.
  $queries = [
    'Projects missing repositories' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT vpp.nid FROM {versioncontrol_project_projects} vpp INNER JOIN {node} n ON n.nid = vpp.nid AND n.created < unix_timestamp() - 60 WHERE vpp.repo_id = 0',
    ],
    'Mismatched full/sandbox repositories' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT vpp.nid, vpp.repo_id, vglr.namespace, fpt.field_project_type_value FROM {versioncontrol_project_projects} vpp INNER JOIN {versioncontrol_gitlab_repositories} vglr ON vglr.repo_id = vpp.repo_id AND vglr.namespace = 'sandbox' INNER JOIN {field_data_field_project_type} fpt ON fpt.entity_id = vpp.nid AND fpt.field_project_type_value = 'full'",
    ],
    'Mismatched sandbox/full repositories' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT vpp.nid, vpp.repo_id, vglr.namespace, fpt.field_project_type_value FROM {versioncontrol_project_projects} vpp INNER JOIN {versioncontrol_gitlab_repositories} vglr ON vglr.repo_id = vpp.repo_id AND vglr.namespace = 'project' INNER JOIN {field_data_field_project_type} fpt ON fpt.entity_id = vpp.nid AND fpt.field_project_type_value = 'sandbox'",
    ],
    'Missing project nodes' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT vpp.nid, vpp.repo_id FROM {versioncontrol_project_projects} vpp LEFT JOIN {node} n ON n.nid = vpp.nid WHERE n.nid IS NULL',
    ],
    'Repositories without projects' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT vr.repo_id, vr.name, from_unixtime(updated) updated FROM {versioncontrol_repositories} vr LEFT JOIN {versioncontrol_project_projects} vpp ON vpp.repo_id = vr.repo_id WHERE vpp.repo_id IS NULL',
    ],
    'Releases without projects' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT fdf_rp.entity_id release_nid, fdf_rp.field_release_project_target_id project_nid, n_r.title FROM {field_data_field_release_project} fdf_rp LEFT JOIN {node} n_p ON n_p.nid = fdf_rp.field_release_project_target_id INNER JOIN {node} n_r ON n_r.nid = fdf_rp.entity_id WHERE n_p.nid IS NULL',
    ],
    'Sandboxes with releases enabled' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT n.nid, n.type, n.title, n.uid, from_unixtime(created) created FROM {field_data_field_project_has_releases} fdf_phr INNER JOIN {field_data_field_project_type} fdf_pt ON fdf_pt.entity_id = fdf_phr.entity_id AND fdf_pt.field_project_type_value = 'sandbox' INNER JOIN {node} n ON n.nid = fdf_phr.entity_id WHERE fdf_phr.field_project_has_releases_value = 1",
    ],
    'Orphaned versioncontrol_git_repositories rows' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT vgr.repo_id, vgr.default_branch FROM {versioncontrol_git_repositories} vgr LEFT JOIN {versioncontrol_repositories} vr ON vr.repo_id = vgr.repo_id WHERE vr.repo_id IS NULL',
    ],
    'Multiple releases per tag/branch' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT group_concat(n.nid) nids, count(1) c, fdf_rp.field_release_project_target_id project, fdf_rv.field_release_version_value version FROM {node} n INNER JOIN {field_data_field_release_project} fdf_rp ON fdf_rp.entity_id = n.nid INNER JOIN {field_data_field_release_version} fdf_rv ON fdf_rv.entity_id = n.nid GROUP BY fdf_rp.field_release_project_target_id, fdf_rv.field_release_version_value HAVING c > 1 ORDER BY NULL',
    ],
    'Missing VCS labels' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT n.nid, n.title FROM {node} n LEFT JOIN {field_data_field_release_vcs_label} fdf_vl ON fdf_vl.entity_id = n.nid WHERE n.type = 'project_release' AND (fdf_vl.field_release_vcs_label_value IS NULL OR fdf_vl.field_release_vcs_label_value = '')",
    ],
    'Missing version' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT n.nid, n.title FROM {node} n LEFT JOIN {field_data_field_release_version} fdf_v ON fdf_v.entity_id = n.nid WHERE n.type = 'project_release' AND (fdf_v.field_release_version_value IS NULL OR fdf_v.field_release_version_value = '')",
    ],
    'Tests missing releases' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT pcp.nid, pcp.environment, pcp.release_nid, pcp.testing, pcp.core_branch FROM {pift_ci_project} pcp LEFT JOIN {node} n ON n.nid = pcp.release_nid WHERE n.nid IS NULL',
    ],
    'Orphaned release files' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT fm.fid, fm.uri FROM {file_managed} fm LEFT JOIN {file_usage} fu ON fu.fid = fm.fid LEFT JOIN {field_collection_item} ci ON ci.item_id = fu.id WHERE fm.uri LIKE 'public://projects/%' AND (fu.id IS NULL OR ci.item_id IS NULL)",
    ],
    'Releases that lost their files' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT q.entity_id, group_concat(fm.fid), group_concat(fm.filename) FROM {file_managed} fm INNER JOIN (SELECT DISTINCT n.entity_id, concat('public://projects/', fdf_pmn.field_project_machine_name_value, '-', fdf_rv.field_release_version_value, '.%') pattern FROM {field_data_field_release_files} n LEFT JOIN {field_data_field_release_file} c ON c.entity_id = n.field_release_files_value INNER JOIN {field_data_field_release_project} fdf_rp ON fdf_rp.entity_id = n.entity_id INNER JOIN {field_data_field_project_machine_name} fdf_pmn ON fdf_pmn.entity_id = fdf_rp.field_release_project_target_id INNER JOIN {field_data_field_release_version} fdf_rv ON fdf_rv.entity_id = n.entity_id WHERE c.entity_id IS NULL) q ON fm.uri LIKE q.pattern GROUP BY q.entity_id",
    ],
    'Files that lost their releases' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT fdf_rf.entity_id, fdf_rf.revision_id, fm.filename, group_concat(fdf_rv.entity_id) release_nid FROM {field_data_field_release_file} fdf_rf INNER JOIN {file_managed} fm ON fm.fid = fdf_rf.field_release_file_fid AND fm.filename NOT LIKE '%-core.tar.gz' AND fm.filename NOT LIKE '%-core.zip' LEFT JOIN {field_data_field_release_files} fdf_nrf ON fdf_nrf.field_release_files_value = fdf_rf.entity_id LEFT JOIN {field_data_field_project_machine_name} fdf_pmn ON fdf_pmn.field_project_machine_name_value = substring_index(fm.filename, '-', 1) LEFT JOIN {field_data_field_release_project} fdf_rp ON fdf_rp.field_release_project_target_id = fdf_pmn.entity_id INNER JOIN {field_data_field_release_version} fdf_rv ON fdf_rv.entity_id = fdf_rp.entity_id AND fdf_rv.field_release_version_value = trim(TRAILING '.' FROM substring(regexp_substr(fm.filename, '-[-x.0-9]+(dev|alpha[0-9]|beta[0-9]|rc[0-9])?\\\\.'), 2)) WHERE fdf_nrf.field_release_files_value IS NULL GROUP BY fdf_rf.entity_id",
    ],
    'In-place-update errors' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT n_from.title from_release, n_to.title to_release, from_unixtime(dipu.requested) requested, from_unixtime(dipu.generated) generated, dipu.generated - dipu.requested delta, dipu.error FROM {drupalorg_in_place_updates} dipu LEFT JOIN {node} n_from ON n_from.nid = dipu.from_release_nid LEFT JOIN {node} n_to ON n_to.nid = dipu.to_release_nid WHERE dipu.error IS NOT NULL ORDER BY dipu.requested',
    ],
    'In-place-updates not generated' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT n_from.title from_release, n_to.title to_release, from_unixtime(dipu.requested) requested, from_unixtime(dipu.generated) generated, dipu.generated - dipu.requested delta, dipu.error FROM {drupalorg_in_place_updates} dipu LEFT JOIN {node} n_from ON n_from.nid = dipu.from_release_nid LEFT JOIN {node} n_to ON n_to.nid = dipu.to_release_nid WHERE dipu.generated IS NULL AND dipu.requested + 120 < unix_timestamp() ORDER BY dipu.requested',
    ],
    'Expired in-place-updates' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT n_from.title from_release, n_to.title to_release, from_unixtime(dipu.requested) requested, from_unixtime(dipu.generated) generated, dipu.generated - dipu.requested delta, from_unixtime(dipu.expiration) expiration, dipu.error FROM {drupalorg_in_place_updates} dipu LEFT JOIN {node} n_from ON n_from.nid = dipu.from_release_nid LEFT JOIN {node} n_to ON n_to.nid = dipu.to_release_nid WHERE dipu.expiration < unix_timestamp() ORDER BY dipu.requested',
    ],
    'Mismatched node status' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT n.nid, nr.nid, n.type, n.title FROM {node} n INNER JOIN {node_revision} nr ON nr.vid = n.vid AND ((n.status = 1 AND nr.status = 0) OR (n.status = 0 AND nr.status = 1))',
    ],
    'Issue index missing issue' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT n.nid, n.title FROM {node} n LEFT JOIN {search_api_db_project_issues} sad_pi ON sad_pi.item_id = n.nid WHERE n.type = 'project_issue' AND n.status = 1 AND sad_pi.item_id IS NULL",
    ],
    'Issue index mismatched version' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT sid_pi.item_id, sid_pi.field_issue_version, fdf_iv.field_issue_version_value FROM {search_api_db_project_issues} sid_pi LEFT JOIN {field_data_field_issue_version} fdf_iv ON fdf_iv.entity_id = sid_pi.item_id WHERE fdf_iv.field_issue_version_value != sid_pi.field_issue_version',
    ],
    'Issue index mismatched status' => [
      'severity' => LogLevel::WARNING,
      'query' => 'SELECT sid_pi.item_id, sid_pi.field_issue_status, fdf_is.field_issue_status_value FROM {search_api_db_project_issues} sid_pi LEFT JOIN {field_data_field_issue_status} fdf_is ON fdf_is.entity_id = sid_pi.item_id WHERE fdf_is.field_issue_status_value != sid_pi.field_issue_status AND sid_pi.changed < unix_timestamp(date_sub(now(), INTERVAL 60 MINUTE))',
    ],
    'Git users without Git role' => [
      'severity' => LogLevel::ERROR,
      'query' => 'SELECT u.uid, u.name FROM {users} u LEFT JOIN {field_data_field_shared_account_for_an_org} fdf_sao ON fdf_sao.entity_id = u.uid LEFT JOIN {users_roles} ur ON ur.uid = u.uid AND ur.rid = 20 WHERE u.status = 1 AND u.git_consent = 1 AND u.git_disabled = 0 AND u.git_username IS NOT NULL AND (fdf_sao.field_shared_account_for_an_org_value != 1 OR fdf_sao.field_shared_account_for_an_org_value IS NULL) AND ur.uid IS NULL',
    ],
    'Projects with releases with releases disabled' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT n.nid, n.title FROM {node} n INNER JOIN {field_data_field_project_type} fdf_pt ON fdf_pt.entity_id = n.nid AND fdf_pt.field_project_type_value = 'full' INNER JOIN {field_data_field_project_has_releases} fdf_phr ON fdf_phr.entity_id = n.nid AND fdf_phr.field_project_has_releases_value = 0 INNER JOIN {field_data_field_release_project} fdf_rp ON fdf_rp.field_release_project_target_id = n.nid GROUP BY n.nid ORDER BY NULL",
    ],
    'Packaging jobs processing for more than an hour' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT job_id, from_unixtime(updated), project_machine_name, reference, message FROM {drupalorg_packaging_job} WHERE status IN ('sent', 'running') AND updated < unix_timestamp(date_sub(now(), INTERVAL 1 HOUR))",
    ],
    'User with duplicate email' => [
      'severity' => LogLevel::ERROR,
      'query' => "SELECT count(1) c, group_concat(u.uid) uids, group_concat(u.name) names FROM {users} u WHERE u.uid > 0 AND u.mail != '' GROUP BY u.mail HAVING c > 1",
    ],
  ];
  $something_failed = FALSE;
  foreach ($queries as $title => $query) {
    if ($result = db_query($query['query'])->fetchAll(PDO::FETCH_ASSOC)) {
      $something_failed = TRUE;
      $message = count($result) . ' ' . dt($title);
      if ($query['severity'] === LogLevel::ERROR) {
        drush_set_error('drupalorg_integrity', $message);
      }
      else {
        drush_log($message, $query['severity']);
      }
      array_unshift($result, array_keys($result[0]));
      drush_print(drush_format_table($result));
    }
  }
  if (!$something_failed) {
    drush_log('🎉');
  }
}

/**
 * Generate release contents hashes for a single release.
 */
function drush_drupalorg_release_hashes($project_argument, $version) {
  // Load the project node.
  if (!($project = project_load($project_argument))) {
    drush_set_error(dt('Could not load project.'));
    return;
  }

  // Load the release node.
  $query = new EntityFieldQuery();
  $result = $query->entityCondition('entity_type', 'node')
    ->propertyCondition('type', project_release_release_node_types())
    ->fieldCondition('field_release_project', 'target_id', $project->nid)
    ->fieldCondition('field_release_version', 'value', $version)
    ->fieldCondition('field_release_build_type', 'value', 'static')
    ->execute();
  if (!isset($result['node'])) {
    drush_set_error(dt('Could not load release.'));
    return;
  }
  $node = node_load(array_keys($result['node'])[0]);
  $wrapper = entity_metadata_wrapper('node', $node);
  if ($node->status != NODE_PUBLISHED) {
    drush_set_error(dt('Release is not published!'));
    return;
  }

  // Prepare directory for saving results.
  $workspace = drush_tempdir();
  $directory = drupalorg_release_hashes_directory($wrapper);
  file_prepare_directory($directory, FILE_CREATE_DIRECTORY);

  // Generate & save hashes for git clone.
  if (empty($node->field_release_vcs_label[LANGUAGE_NONE][0]['value'])) {
    watchdog('drupalorg_hashes', 'Label missing for !release', [
      '!release' => l($node->title, 'node/' . $node->nid),
    ], WATCHDOG_NOTICE);
  }
  else {
    try {
      if (!drush_shell_cd_and_exec($workspace, 'git clone --depth 1 --branch %s %s clone', $node->field_release_vcs_label[LANGUAGE_NONE][0]['value'], versioncontrol_project_repository_load($project->nid)->remoteUrl())) {
        throw new Exception(t('Could not clone !release archive: <pre>@output</pre>', [
          '!release' => l($node->title, 'node/' . $node->nid),
          '@output' => implode("\n", drush_shell_exec_output()),
        ]));
      }
      drupalorg_save_signed_data(drupalorg_get_file_hashes($workspace . '/clone', "-not -regex '\./\.git/.*'"), $directory . '/contents-sha256sums.csig', 'Release Git tag contents hashes for ' . $node->title);
    }
    catch (Exception $e) {
      watchdog('drupalorg_hashes', 'Exception in processing git clone for !release release: @message', [
        '!release' => l($node->title, 'node/' . $node->nid),
        '@message' => $e->getMessage(),
      ], WATCHDOG_ERROR);
    }
  }

  // Generate & save hashes for packaged files.
  try {
    drupalorg_extract_release($wrapper, $workspace . '/package');
    drupalorg_save_signed_data(drupalorg_get_file_hashes($workspace . '/package'), $directory . '/contents-sha256sums-packaged.csig', 'Release package contents hashes for ' . $node->title);
  }
  catch (Exception $e) {
    watchdog('drupalorg_hashes', 'Exception in processing packaged !release release: @message', [
      '!release' => l($node->title, 'node/' . $node->nid),
      '@message' => $e->getMessage(),
    ], WATCHDOG_ERROR);
  }
}

/**
 * Saves a string to the specified destination with signing and cache clearing.
 *
 * @param $data
 *   A string containing the contents of the file.
 * @param $destination
 *   A string containing the destination location. This must be a stream wrapper
 *   URI.
 * @param string $reason
 *   For the audit log, reason a signing request is being made.
 */
function drupalorg_save_signed_data($data, $destination, $reason) {
  static $files_url_prefix;

  if (!isset($files_url_prefix)) {
    $files_url_prefix = '/^' . preg_quote(file_create_url('public://'), '/') . '/';
  }

  // Save the file.
  file_unmanaged_save_data(drupalorg_sign_data($data, $reason), $destination, FILE_EXISTS_REPLACE);

  // Always purge the CDN, 404s can be cached.
  drupalorg_crosssite_fastly_purge_url(preg_replace($files_url_prefix, 'https://updates.drupal.org/', file_create_url($destination)));
}

/**
 * Signs a string of data.
 *
 * @param string $data
 *   The data to sign.
 * @param string $reason
 *   For the audit log, reason a signing request is being made.
 * @param \DateTimeImmutable $valid_through
 *   If provided, filled with the valid-through date of the signed message.
 *
 * @return string
 *   The signed data.
 */
function drupalorg_sign_data($data, $reason, \DateTimeImmutable &$valid_through = NULL) {
  static $n = 0;
  static $reply_to;
  static $sender;
  static $stomp_client;
  static $verifier;

  // If not valid UTF8, try converting it.
  if (!drupal_validate_utf8($data)) {
    $data = utf8_encode($data);
  }

  if ($broker = variable_get('drupalorg_signing_broker')) {
    if (!isset($reply_to)) {
      $n += 1;
      $sender = gethostname() . '-' . getmypid() . '-' . $n;
      $reply_to = '/temp-queue/SigningResponses-' . $sender;
      $connection = new Stomp\Network\Connection($broker['uri']);
      $connection->setWriteTimeout($broker['write_timeout']);
      $client = new Stomp\Client($connection);
      $client->setLogin($broker['username'], $broker['password']);
      $stomp_client = new Stomp\StatefulStomp($client);
      $stomp_client->subscribe($reply_to, NULL, 'client-individual');
      $verifier = new Verifier(file_get_contents(drupal_get_path('module', 'drupalorg') . '/signing-key/root.pub'));
    }

    $correlation_id = Stomp\Util\IdGenerator::generateId();
    $stomp_message = new Stomp\Transport\Message(gzencode(json_encode([
      'signable-payload' => $data,
      'sender' => $sender,
      'reason' => $reason,
    ])), [
      'reply-to' => $reply_to,
      'correlation-id' => $correlation_id,
      'type' => 'application/json',
      'encoding' => 'gzip',
    ]);
    if ($code = json_last_error()) {
      throw new Exception(t('Failed to encode JSON, error @code', ['@code' => $code]));
    }
    try {
      if (!$stomp_client->send('SigningRequests', $stomp_message)) {
        throw new Exception(t('Failed to send message to @broker', ['@broker' => $broker['uri']]));
      }
    }
    catch (\Stomp\Exception\ConnectionException $e) {
      // Reset connection so that this has a chance of succeeding next time.
      unset($reply_to);
      throw $e;
    }

    // Get result.
    $n = 0;
    while (($frame = $stomp_client->read()) === FALSE) {
      usleep(500);
      $n += 1;
      if ($n > 100) {
        break;
      }
    }
    if ($frame === FALSE) {
      throw new Exception(t('Timeout from @broker', ['@broker' => $broker['uri']]));
    }
    watchdog('drupalorg_signing', 'Got signing response after @n tries.', ['@n' => $n], WATCHDOG_DEBUG);

    // Verify correlation ID.
    if ($correlation_id != $frame->getHeaders()['correlation-id']) {
      $stomp_client->nack($frame);
      throw new Exception(t('Mismatched coorelation ID from @broker', ['@broker' => $broker['uri']]));
    }
    $stomp_client->ack($frame);
    $signed_data = $frame->getBody();

    Stomp\Util\IdGenerator::releaseId($correlation_id);
  }
  else {
    // If there is not a broker configuration, use test keys and sign locally
    // for testing. See
    // https://github.com/drupal/php-signify#chaining-with-csig
    watchdog('drupalorg_signing', 'Using development signing with temporary keys.', [], WATCHDOG_INFO);

    if (!isset($verifier)) {
      $verifier = new Verifier(file_get_contents(drupal_get_path('module', 'drupalorg') . '/temporary-keys/root.pub'));
    }

    // Sign with the build key.
    $process = proc_open('signify-openbsd -S -e -s ' . drupal_get_path('module', 'drupalorg') . "/temporary-keys/intermediate.sec -m - -x - -c 'verify with intermediate.pub'", [
      0 => ['pipe', 'r'],
      1 => ['pipe', 'w'],
      2 => ['pipe', 'w'],
    ], $pipes);
    if (is_resource($process)) {
      fwrite($pipes[0], $data);
      fclose($pipes[0]);
      $signed_data = stream_get_contents($pipes[1]);
      fclose($pipes[1]);
      $stderr = stream_get_contents($pipes[2]);
      fclose($pipes[2]);
      $exit_status = proc_close($process);
      if ($exit_status !== 0) {
        throw new Exception(t('Non-zero exit status!'));
      }
    }
    else {
      throw new Exception(t('Could not open process!'));
    }

    // Prepend intermediate key.
    $signed_data = file_get_contents(drupal_get_path('module', 'drupalorg') . '/temporary-keys/intermediate.xpub') . $signed_data;
  }

  // Verify the signed data.
  if ($verifier->verifyCsigMessage($signed_data) !== $data) {
    throw new Exception(t('Signed data does not match original!'));
  }

  // Extract the valid thorugh time from the 3rd line of the signed data.
  $valid_through = \DateTimeImmutable::createFromFormat('Y-m-d', explode("\n", $signed_data, 4)[2], new \DateTimeZone('UTC'));

  return $signed_data;
}

/**
 * Extract a release’s archive for use in drush commands.
 *
 * @param EntityDrupalWrapper $release_wrapper
 *   The entity metadata wrapper for the release.
 * @param string $destination
 *   Where to extract the release.
 */
function drupalorg_extract_release(EntityDrupalWrapper $release_wrapper, $destination) {
  $file = NULL;
  foreach ($release_wrapper->field_release_files as $possible_file) {
    if (preg_match('#(?<!-core)\.tar\.gz$#', $possible_file->field_release_file->value()['filename'])) {
      $file = $possible_file->field_release_file->value();
    }
  }
  if (is_null($file)) {
    throw new Exception(t('Could not find archive.'));
  }
  if (!mkdir($destination)) {
    throw new Exception(t('Could not create extraction directory.'));
  }
  if (!drush_shell_cd_and_exec($destination, 'tar xf %s --strip-components=1', drupal_realpath($file['uri']))) {
    throw new Exception(t('Could not extract archive: <pre>@output</pre>', [
      '@output' => implode("\n", drush_shell_exec_output()),
    ]));
  }
}

/**
 * Get sha256sum --tag hashes of files.
 *
 * @param string $directory
 *   Directory to find files in.
 * @param string $extra_find_args
 *   Extra arguments to pass to the find command.
 */
function drupalorg_get_file_hashes($directory, $extra_find_args = '') {
  if (!drush_shell_cd_and_exec($directory, 'find . -type f ' . $extra_find_args . " -printf '%P\\0' | xargs -0 -n 50 -P 2 sha256sum --tag")) {
    throw new Exception(t('Could not get checksums: <pre>@output</pre>', [
      '@output' => implode("\n", drush_shell_exec_output()),
    ]));
  }

  return implode("\n", drush_shell_exec_output()) . "\n";
}

/**
 * Re-sign csig files with a new intermediate key.
 */
function drush_drupalorg_resign_files() {
  $verifier = new Verifier(file_get_contents(drupal_get_path('module', 'drupalorg') . '/signing-key/root.pub'));
  $file_path_prefix = '/^' . preg_quote(drupal_realpath('public://'), '/') . '/';

  foreach (explode("\0", trim(file_get_contents('php://stdin'))) as $filename) {
    $path = 'public:/' . preg_replace($file_path_prefix, '', realpath($filename));
    drush_log(dt('Processing @path', ['@path' => $path]));

    $contents = file_get_contents($path);
    if ($contents === FALSE) {
      drush_log(dt('Could not read file @filename', ['@filename' => $filename]), LogLevel::ERROR);
      continue;
    }

    try {
      drupalorg_save_signed_data($verifier->verifyCsigMessage($contents), $path, 'Re-signing ' . $filename);
    }
    catch (\Exception $e) {
      watchdog('drupalorg_hashes', 'Exception in re-signing @filename: @message', [
        '@filename' => $filename,
        '@message' => $e->getMessage(),
      ], WATCHDOG_ERROR);
    }
  }
}

/**
 * Send a packaged release to rugged for TUF.
 */
function drush_drupalorg_release_send_rugged($project_argument) {
  // Load the project node.
  if (!($project = project_load($project_argument))) {
    drush_set_error(dt('Could not load project.'));
    return;
  }
  $release_category = 'current';

  // Load the release nodes.
  if ($project_argument !== 'drupal') {
    $query = (new EntityFieldQuery())->entityCondition('entity_type', 'node')
      ->propertyCondition('type', project_release_release_node_types())
      ->propertyCondition('status', NODE_PUBLISHED)
      ->fieldCondition('field_release_project', 'target_id', $project->nid)
      ->fieldCondition('field_release_category', 'value', $release_category);
    if ($version = drush_get_option('version')) {
      $query->fieldCondition('field_release_version', 'value', $version);
    }
    $result = $query->execute();
    if (!isset($result['node'])) {
      drush_set_error(dt('Could not load releases.'));
      return;
    }
  }

  // Set up TUF targets directory structure.
  if (!($namespace = project_composer_get_project_namespace($project->nid, $release_category, FALSE))) {
    drush_set_error(dt('No namespace.'));
    return;
  }
  $suffix = (new DateTimeImmutable('now', new DateTimeZone('UTC')))->format('c') . '_' . getmypid();
  drush_log(dt('Building @suffix for @project_argument', ['@suffix' => $suffix, '@project_argument' => $project_argument]));

  try {
    $building_directory = 'post-to-tuf/building_' . $suffix;
    $package_directory = $building_directory . '/drupal/' . $namespace;

    $files_to_copy = [];
    if ($project_argument !== 'drupal') {
      foreach (node_load_multiple(array_keys($result['node'])) as $release_node) {
        $release_wrapper = entity_metadata_wrapper('node', $release_node);

        // Locate the zip archive.
        $file = NULL;
        foreach ($release_wrapper->field_release_files as $possible_file) {
          if (preg_match('#(?<!-core)\.zip$#', $possible_file->field_release_file->value()['filename'])) {
            $file = $possible_file->field_release_file->value();
          }
        }
        if (is_null($file)) {
          drush_set_error(dt('Could not find @version archive.', ['@version' => $release_wrapper->field_release_version->value()]));
          continue;
        }
        if (($package_path = drupal_realpath($file['uri'])) === FALSE) {
          drush_set_error(dt('Could not find @version file.', ['@version' => $release_wrapper->field_release_version->value()]));
          continue;
        }

        // Normalize version number.
        if (($version = _project_composer_format_version($release_node)) === FALSE) {
          drush_log(dt('Version @version is not normalizable.', ['@version' => $release_wrapper->field_release_version->value()]));
          continue;
        }
        if ($release_wrapper->field_release_version_extra->value() === 'dev') {
          // -dev must be the suffix to normalize properly.
          $version = preg_replace('/^dev-/', '', $version) . '-dev';
        }
        $normalized_version = (new \Composer\Semver\VersionParser())->normalize($version);

        // Copy package to VENDOR/PACKAGE_NAME/NORMALIZED_VERSION.
        $files_to_copy[$package_directory . '/' . $normalized_version] = $package_path;
      }
    }

    // Copy Composer metadata.
    $metadata_prefix = '/' . variable_get('project_composer_metadata_root', 'packages') . '/' . project_composer_core_version_string_from_release_category($release_category);
    $composer_metadata_directory = drupal_realpath(file_build_uri(''));
    $files_directory = '/' . (new DrupalPublicStreamWrapper())->getDirectoryPath();
    $project_metadata_path = $metadata_prefix . '/p2/drupal/' . $namespace;
    $files_to_copy[$building_directory . $files_directory . $project_metadata_path . '.json'] = $composer_metadata_directory . $project_metadata_path . '.json';
    if ($project_argument !== 'drupal') {
      $files_to_copy[$building_directory . $files_directory . $project_metadata_path . '~dev.json'] = $composer_metadata_directory . $project_metadata_path . '~dev.json';
    }
    if (!drush_get_option('skip-packages-json', FALSE)) {
      $files_to_copy[$building_directory . '/packages.json'] = $composer_metadata_directory . $metadata_prefix . '/packages.json';
    }

    // Upload files.
    list($username, $hostname) = explode('@', variable_get('drupalorg_rugged_sftp', 'rugged@localhost'));
    $sftp = new SFTP($hostname);
    if (!$sftp->login($username, phpseclib3\Crypt\PublicKeyLoader::load(file_get_contents(variable_get('drupalorg_rugged_sftp_key_path', '~/.ssh/id_ed25519-rugged'))))) {
      throw new \Exception('SFTP login failed');
    }
    $sftp->mkdir($package_directory, -1, TRUE);
    $sftp->mkdir(dirname($building_directory . $files_directory . $project_metadata_path), -1, TRUE);
    foreach ($files_to_copy as $remote => $local) {
      $sftp->put($remote, $local, SFTP::SOURCE_LOCAL_FILE);
    }
    $sftp->rename($building_directory, 'post-to-tuf/tuf_ready_' . $suffix);
  }
  catch (\Exception $e) {
    watchdog('drupalorg_send_rugged', 'Exception in sending @project_argument: @message', [
      '@project_argument' => $project_argument,
      '@message' => $e->getMessage(),
    ], WATCHDOG_ERROR);
    return;
  }
}

/**
 * Publish an update to the announcements.json API.
 */
function drush_drupalorg_update_announcements_json() {
  $body_instance = field_info_instance('node', 'body', 'post');
  $items = [];

  $result = (new EntityFieldQuery())->entityCondition('entity_type', 'node')
    ->propertyCondition('type', 'post')
    ->propertyCondition('status', NODE_PUBLISHED)
    ->fieldCondition('og_group_ref', 'target_id', 3223030)
    ->propertyOrderBy('created', 'DESC')
    ->range(0, 100)
    ->execute();
  foreach (node_load_multiple(array_keys($result['node'])) as $node) {
    $node_wrapper = entity_metadata_wrapper('node', $node);
    $items[] = [
      'id' => $node->nid,
      'title' => $node->title,
      'content_html' => empty($node->body) ? '' : ($node_wrapper->body->summary->value() ?: text_summary($node_wrapper->body->value->value(), $node_wrapper->body->format->value(), $body_instance['display']['teaser']['settings']['trim_length'])),
      'url' => url('node/' . $node->nid, ['absolute' => TRUE]),
      'date_modified' => date('c', $node->changed),
      'date_published' => date('c', $node->created),
      '_drupalorg' => [
        'featured' => (bool) $node->sticky,
        'version' => empty($node->field_announcement_version) ? '*' : $node->field_announcement_version[LANGUAGE_NONE][0]['value'],
      ]
    ];
  }

  // https://www.jsonfeed.org/version/1.1/
  // Domain is hard-coded so test fixtures can be generated from dev & staging.
  $drupal_org_url = 'https://www.drupal.org';
  $filename = drush_get_option('test', FALSE) ? 'announcements-this-is-only-a-test.json' : 'announcements.json';
  $json_feed = [
    'version' => 'https://jsonfeed.org/version/1.1',
    'title' => t('Drupal Announcements Feed'),
    'home_page_url' => $drupal_org_url,
    'feed_url' => $drupal_org_url . '/' . $filename,
    'favicon' => $drupal_org_url . '/favicon.ico',
    'items' => $items,
  ];

  $json = drupal_json_encode($json_feed);

  if (!drush_confirm(dt('New contents to be written to @filename:', ['@filename' => $filename]) . "\n" . $json . "\n" . dt('Do you want to continue?'))) {
    return;
  }

  $directory = 'public://announcements-feed';
  file_prepare_directory($directory, FILE_CREATE_DIRECTORY);
  file_unmanaged_save_data($json, $directory . '/' . $filename, FILE_EXISTS_REPLACE);
  drupalorg_crosssite_fastly_purge_url(url($filename, ['absolute' => TRUE]));
}

/**
 * Publish an update to the psa.json API.
 */
function drush_drupalorg_update_psa_json($nids) {
  $output = [];

  foreach (node_load_multiple(explode(',', $nids)) as $node) {
    if ($node->type !== 'sa') {
      return drush_set_error('Not SA', dt('Not a valid SA.'));
    }

    $project = node_load($node->field_project[LANGUAGE_NONE][0]['target_id']);
    $output[] = [
      'title' => $node->title,
      'link' => url('node/' . $node->nid, ['absolute' => TRUE]),
      'project' => $project->field_project_machine_name[LANGUAGE_NONE][0]['value'],
      'type' => str_replace('project_', '', $project->type),
      'is_psa' => $node->field_is_psa[LANGUAGE_NONE][0]['value'],
      'insecure' => drupalorg_get_insecure_strings($project->nid),
      'pubDate' => date('c', $node->changed),
    ];
  }

  $json = drupal_json_encode($output);

  $filename = drush_get_option('test', FALSE) ? 'psa-this-is-only-a-test.json' : 'psa.json';
  if (!drush_confirm(dt('New contents to be written to @filename:', ['@filename' => $filename]) . "\n" . $json . "\n" . dt('Do you want to continue?'))) {
    return;
  }

  file_unmanaged_save_data($json, 'public://release-history/' . $filename, FILE_EXISTS_REPLACE);
  drupalorg_crosssite_fastly_purge_url('https://updates.drupal.org/' . $filename);
}

/**
 * Get version numbers of releases for a project that are marked as insecure.
 *
 * @param int $nid
 *   Project node ID.
 *
 * @return array
 *   Array of version strings.
 */
function drupalorg_get_insecure_strings($nid) {
  static $insecure_term;

  if (!isset($insecure_term)) {
    $insecure_term = array_values(taxonomy_get_term_by_name('Insecure', 'vocabulary_7'))[0];
  }

  $query = new EntityFieldQuery();
  $result = $query->entityCondition('entity_type', 'node')
    ->entityCondition('bundle', project_release_release_node_types())
    ->fieldCondition('field_release_project', 'target_id', $nid)
    ->fieldCondition('field_release_build_type', 'value', 'static')
    ->fieldCondition('taxonomy_vocabulary_7', 'tid', $insecure_term->tid)
    ->execute();
  if (!isset($result['node'])) {
    return [];
  }
  $versions = [];
  foreach (node_load_multiple(array_keys($result['node'])) as $insecure_release) {
    $versions[$insecure_release->field_release_version[LANGUAGE_NONE][0]['value']] = TRUE;
  }

  return array_keys($versions);
}

/**
 * Generate a quasi-patch for a single release.
 */
function drush_drupalorg_create_quasi_patch($from_release_nid, $to_release_nid) {
  try {
    $releases = node_load_multiple([$from_release_nid, $to_release_nid]);
    $from_release_wrapper = entity_metadata_wrapper('node', $releases[$from_release_nid]);
    $to_release_wrapper = entity_metadata_wrapper('node', $releases[$to_release_nid]);

    $workspace = drush_tempdir();

    // Extract the archives and get file hashes.
    drupalorg_extract_release($from_release_wrapper, $workspace . '/from');
    $from_checksums = [];
    foreach ((new ChecksumList(drupalorg_get_file_hashes($workspace . '/from'), TRUE)) as $checksum) {
      $from_checksums[$checksum->filename] = $checksum->hex_hash;
    }
    drupalorg_extract_release($to_release_wrapper, $workspace . '/to');
    $to_checksums = [];
    foreach ((new ChecksumList(drupalorg_get_file_hashes($workspace . '/to'), TRUE)) as $checksum) {
      $to_checksums[$checksum->filename] = $checksum->hex_hash;
    }

    // Create the quasi-diff.
    mkdir($workspace . '/result');

    // Write manifest of files that are no longer present.
    file_put_contents($workspace . '/result/DELETION_MANIFEST.txt', implode("\n", array_keys(array_diff_key($from_checksums, $to_checksums))) . "\n");

    // Copy files that are new or changed.
    mkdir($workspace . '/result/files');
    $process = proc_open('cpio --pass-through --make-directories --preserve-modification-time -0 ' . escapeshellarg($workspace . '/result/files'), [
      0 => ['pipe', 'r'],
      1 => ['pipe', 'w'],
      2 => ['pipe', 'w'],
    ], $pipes, $workspace . '/to');
    if (is_resource($process)) {
      fwrite($pipes[0], implode("\0", array_keys(array_diff_assoc($to_checksums, $from_checksums))));
      fclose($pipes[0]);
      $data = stream_get_contents($pipes[1]);
      fclose($pipes[1]);
      $stderr = stream_get_contents($pipes[2]);
      fclose($pipes[2]);
      $exit_status = proc_close($process);
      if ($exit_status !== 0) {
        throw new Exception(t('Non-zero exit status from cpio: <pre>@output</pre>', [
          '@output' => $stderr,
        ]));
      }
    }
    else {
      throw new Exception(t('Could not open process!'));
    }

    // Archive the result.
    mkdir($workspace . '/archive');
    $destination = drupalorg_in_place_update_file_path($from_release_wrapper, $to_release_wrapper);
    $filename = basename($destination);
    if (!drush_shell_cd_and_exec($workspace . '/result', 'zip -rq ../archive/' . $filename . ' .')) {
      throw new Exception(t('Could not create zip archive: <pre>@output</pre>', [
        '@output' => implode("\n", drush_shell_exec_output()),
      ]));
    }

    // Sign the archive.
    file_put_contents($workspace . '/archive/' . $filename . '.csig', drupalorg_sign_data(drupalorg_get_file_hashes($workspace . '/archive'), 'In place update artifact for ' . $from_release_wrapper->label() . ' to ' . $to_release_wrapper->label(), $valid_through));

    // Move into place & purge cache.
    file_unmanaged_move($workspace . '/archive/' . $filename, $destination, FILE_EXISTS_REPLACE);
    file_unmanaged_move($workspace . '/archive/' . $filename . '.csig', $destination . '.csig', FILE_EXISTS_REPLACE);
    drupalorg_crosssite_fastly_purge_url(preg_replace('#/files#', '', file_create_url($destination)));
    drupalorg_crosssite_fastly_purge_url(preg_replace('#/files#', '', file_create_url($destination . '.csig')));

    // Record that it was generated.
    db_update('drupalorg_in_place_updates')
      ->condition('from_release_nid', $from_release_wrapper->getIdentifier())
      ->condition('to_release_nid', $to_release_wrapper->getIdentifier())
      ->fields([
        'generated' => time(),
        'expiration' => $valid_through->getTimestamp(),
      ])
      ->execute();
  }
  catch (Exception $e) {
    $message = $e->getMessage();
    watchdog('drupalorg_in_place_updates', 'Exception in creating in place update @from → @to: @message', [
      '@from' => $from_release_nid,
      '@to' => $to_release_nid,
      '@message' => $message,
    ], WATCHDOG_ERROR);
    db_update('drupalorg_in_place_updates')
      ->condition('from_release_nid', $from_release_wrapper->getIdentifier())
      ->condition('to_release_nid', $to_release_wrapper->getIdentifier())
      ->fields([
        'error' => $message,
      ])
      ->execute();
  }
}

/**
 * Updates the status of a job when jenkins starts handling it.
 */
function drush_drupalorg_packaging_status($job_id, $status, $message = '') {
  /* @var $packaging_job \DrupalorgPackagingJob */
  $packaging_job = entity_load_single('drupalorg_packaging_job', $job_id);

  $packaging_job->status = $status;
  $packaging_job->message = $message;
  try {
    $packaging_job->save();
  }
  catch (Exception $e) {
    watchdog('drupalorg_packaging', 'Problem updating packaging job for @project @reference: @message', [
      '@project' => $packaging_job->project_machine_name,
      '@reference' => $packaging_job->reference,
      '@message' => $e->getMessage(),
    ], WATCHDOG_ERROR);

  }
}

/**
 * Updates the status of a job when jenkins starts handling it.
 */
function drush_drupalorg_packaging_resend() {

  // TODO: connect to jenkins api and find DRUPALORG_JOB_STATUS_SENT jobs
  // That jenkins doesnt know about.

  $query = new EntityFieldQuery();
  $query->entityCondition('entity_type', 'drupalorg_packaging_job')
    ->propertyCondition('status', [DRUPALORG_JOB_STATUS_QUEUE, DRUPALORG_JOB_STATUS_STUCK], 'IN');
  $result = $query->execute();

  if (isset($result['drupalorg_packaging_job'])) {
    foreach (entity_load('drupalorg_packaging_job', array_keys($result['drupalorg_packaging_job'])) as $job) {
      if ($job->status == DRUPALORG_JOB_STATUS_STUCK ) {
        // set it back to 'queued' if it was previously stuck and let it try again.
        $job->status = DRUPALORG_JOB_STATUS_QUEUE;
        $job->save();
        $job->queueJob();
      } else {
        $job->queueJob();
      }
    }
  }
}

/**
 * Re-pathauto a tree of documentation.
 */
function drush_drupalorg_repath_docs($nid) {
  $nids = [$nid];

  while (!empty($nids)) {
    $nodes = node_load_multiple($nids);
    $nids = [];
    foreach ($nodes as $node) {
      $existing = drupal_get_path_alias('node/' . $node->nid);
      if (!empty($node->path['pathauto'])) {
        $new = pathauto_node_update_alias($node, 'return');
        if ($existing !== $new) {
          // Update the alias.
          drush_log(dt('@existing → @new', [
            '@existing' => $existing,
            '@new' => $new,
          ]));
          pathauto_node_update_alias($node, 'bulkupdate');
        }
        if ($node->type === 'guide') {
          // Traverse pages & guides.
          $result = (new EntityFieldQuery())->entityCondition('entity_type', 'node')
            ->fieldCondition('og_group_ref_documentation', 'target_id', $node->nid)
            ->execute();
          if (!empty($result['node'])) {
            $nids += array_keys($result['node']);
          }
        }
      }
      else {
        // Automatic alias was disabled.
        drush_log(dt('Skipping @existing', ['@existing' => $existing]));
      }
    }
  }
}

/**
 * Registers a project at Tugboat.
 */
function drush_drupalorg_tugboat_register_project($project_namespace, $project_name) {
  try {
    $response = TugboatApi::registerRepository($project_namespace, $project_name);
    drush_log(dt('Response is: !response', [
      '!response' => $response,
    ]), LogLevel::OK);
  }
  catch (\GuzzleHttp\Exception\BadResponseException $e) {
    drush_log(dt('Bad response: @message', [
      '@message' => (string) $e->getResponse()->getBody(),
    ]), LogLevel::ERROR);
  }
  catch (Exception $e) {
    drush_log(dt('Exception: @message', [
      '@message' => $e->getMessage(),
    ]), LogLevel::ERROR);
  }
}

/**
 * Builds a base preview for a Tugboat project.
 */
function drush_drupalorg_tugboat_build_base_preview($project_name, $branch) {
  $tugboat_configuration = TugboatPreview::fetchConfiguration($project_name, $branch);

  if (empty($tugboat_configuration)) {
    drush_log(dt('Could not fetch Tugboat configuration for such project and branch.'), LogLevel::ERROR);
    return;
  }

  $body = [
    'ref' => $branch,
    'repo' => TugboatApi::getRepositoryId('project/' . $project_name),
    'config' => $tugboat_configuration,
    'name' => $branch,
    'anchor' => TRUE,
    'anchor_type' => 'branch',
    'type' => 'branch',
  ];

  try {
    $response = TugboatApi::getClient()->request('POST', '/v3/previews', [
      'headers' => ['Content-Type' => 'application/json'],
      'body' => json_encode($body),
    ]);
    drush_log(dt('Response is: !response', [
      '!response' => $response->getBody()->__toString(),
    ]), LogLevel::OK);
  }
  catch (Exception $e) {
    drush_log(dt('Exception: @message', ['@message' => $e->getMessage()]), LogLevel::ERROR);
  }
}

/**
 * Marks releases for projects with revoked security advisory coverage as insecure.
 */
function drush_drupalorg_mark_revoked_project_releases_insecure() {
  $revoked_projects = (new EntityFieldQuery())->entityCondition('entity_type', 'node')
    ->entityCondition('bundle', project_project_node_types())
    ->fieldCondition('field_security_advisory_coverage', 'value', 'revoked')
    ->execute();
  $revoked_project_nids = array_keys($revoked_projects['node']);
  $all_releases = (new EntityFieldQuery())->entityCondition('entity_type', 'node')
    ->entityCondition('bundle', project_release_release_node_types())
    ->fieldCondition('field_release_project', 'target_id', $revoked_project_nids, 'IN')
    ->execute();
  $insecure_releases = (new EntityFieldQuery())->entityCondition('entity_type', 'node')
    ->entityCondition('bundle', project_release_release_node_types())
    ->fieldCondition('field_release_project', 'target_id', $revoked_project_nids)
    ->fieldCondition('taxonomy_vocabulary_7', 'tid', 188131)
    ->execute();
  $release_nids = array_diff(array_keys($all_releases['node']), array_keys($insecure_releases['node']));
  if (empty($release_nids)) {
    drush_log(dt('There is nothing to do.'));
    return;
  }
  if (!drush_confirm(dt('@count releases will be marked as insecure', ['@count' => count($release_nids)]))) {
    return;
  }
  foreach (node_load_multiple(array_chunk($release_nids, 50)[0]) as $node) {
    drush_log(dt('Processing @title', ['@title' => $node->title]));
    $node->taxonomy_vocabulary_7[LANGUAGE_NONE][] = ['tid' => 188131];
    $node->revision = TRUE;
    $node->log = t('Project’s security advisory coverage was revoked.');
    node_save($node);
  }
  if (count($release_nids) > 50) {
    drush_log(dt('Only a batch of 50 was processed. Run again for another batch.'));
  }
}

/**
 * Updates files/project_analysis/allprojects.tsv for
 * https://www.drupal.org/project/project_analysis
 */
function drush_drupalorg_update_project_analysis_data() {
  $path = 'public://project_analysis/allprojects.tsv';
  $stream = fopen(drupal_realpath($path), 'w');
  db_query('SET SESSION group_concat_max_len = 65535');
  $result = db_query("SELECT
    fdfpmn.field_project_machine_name_value,
    pcnm.package_namespace,
    REGEXP_REPLACE(SUBSTRING_INDEX(fdf_rvl.field_release_vcs_label_value,'8.x-',-1), '.x','.x-dev') as Composer,
    n.type,
    (SELECT sum(count) FROM {project_usage_week_release} puwr INNER JOIN {field_data_field_release_project} fdf_rp ON fdf_rp.entity_id = puwr.nid INNER JOIN {field_data_field_release_category} fdf_rc ON fdf_rc.entity_id = fdf_rp.entity_id AND fdf_rc.field_release_category_value = 'current' WHERE fdf_rp.field_release_project_target_id = prsv.nid AND puwr.timestamp = (SELECT max(timestamp) FROM {project_usage_week_release})),
    IF (fdtv44.taxonomy_vocabulary_44_tid != 13032, '', 'Abandoned'),
    IF (fdtv46.taxonomy_vocabulary_46_tid != 9994, '', 'Obsolete'),
    group_concat(DISTINCT u.name)
  FROM {project_release_supported_versions} prsv
    INNER JOIN {field_data_field_project_machine_name} fdfpmn ON fdfpmn.entity_id = prsv.nid
    LEFT JOIN (SELECT n.nid, fdf_rp.field_release_project_target_id project_nid, fdf_rv.field_release_version_value version FROM {node} n INNER JOIN {field_data_field_release_project} fdf_rp ON fdf_rp.entity_id = n.nid INNER JOIN {field_data_field_release_version} fdf_rv ON fdf_rv.entity_id = fdf_rp.entity_id WHERE n.status = 1) dev_release ON dev_release.project_nid = prsv.nid AND dev_release.version IN (concat(substring_index(prsv.branch, '.', 1), '.x-dev'), concat(prsv.branch, 'x-dev'))
    INNER JOIN {node} n on n.nid = prsv.nid AND n.status = 1 AND n.type IN ('project_module', 'project_theme')
    INNER JOIN {field_data_field_release_category} fdf_rc ON fdf_rc.entity_id = prsv.latest_release AND fdf_rc.field_release_category_value = 'current'
    INNER JOIN {project_composer_namespace_map} pcnm ON pcnm.project_nid = fdfpmn.entity_id AND pcnm.component_name = fdfpmn.field_project_machine_name_value AND pcnm.category = 'current'
    LEFT JOIN {field_data_taxonomy_vocabulary_44} fdtv44 on prsv.nid = fdtv44.entity_id
    LEFT JOIN {field_data_taxonomy_vocabulary_46} fdtv46 on prsv.nid = fdtv46.entity_id
    INNER JOIN {field_data_field_release_vcs_label} fdf_rvl ON fdf_rvl.entity_id = coalesce(dev_release.nid, prsv.latest_release)
    INNER JOIN {field_data_field_security_advisory_coverage} fdf_sac ON fdf_sac.entity_id = prsv.nid AND fdf_sac.field_security_advisory_coverage_value != 'revoked'
    INNER JOIN {versioncontrol_project_projects} vpp ON vpp.nid = prsv.nid
    LEFT JOIN {versioncontrol_auth_account} vaa ON vaa.repo_id = vpp.repo_id AND vaa.access != 0
    LEFT JOIN {users} u ON u.uid = vaa.uid AND u.status = 1
  WHERE prsv.supported = 1
  GROUP BY prsv.nid, prsv.branch
  ORDER BY NULL", [], ['fetch' => PDO::FETCH_ASSOC]);
  foreach ($result as $row) {
    fputcsv($stream, $row, "\t");
  }
  fclose($stream);
  drupalorg_crosssite_fastly_purge_url(file_create_url($path));
}

/**
 * Marks contrib releases with stale core compatibility as unsupported.
 */
function drush_drupalorg_unsupport_stale_core_compatibility($core_constraint) {
  // Get matching core versions.
  $check_versions = [];
  foreach (project_release_get_releases_by_version(3060) as $core_release) {
    if (\Composer\Semver\Semver::satisfies($core_release->field_release_version[LANGUAGE_NONE][0]['value'], $core_constraint)) {
      $check_versions[] = $core_release->field_release_version[LANGUAGE_NONE][0]['value'];
    }
  }

  // Get possible core version requirements of supported current releases.
  $incompatible_core_version_requirements = [];
  $result = db_query("SELECT DISTINCT pcc.core_version_requirement FROM {project_release_supported_versions} prsv INNER JOIN {node} n ON n.nid = prsv.nid AND n.status = 1 AND n.type != 'project_core' INNER JOIN {field_data_field_release_category} fdf_rc ON fdf_rc.entity_id = prsv.latest_release AND fdf_rc.field_release_category_value = 'current' INNER JOIN {project_composer_component} pcc ON pcc.release_nid = fdf_rc.entity_id AND pcc.component_role = 'primary' WHERE supported = 1");
  foreach ($result as $row) {
    try {
      $compatible = FALSE;
      foreach ($check_versions as $check_version) {
        if (\Composer\Semver\Semver::satisfies($check_version, $row->core_version_requirement)) {
          $compatible = TRUE;
          break;
        }
      }
      if (!$compatible) {
        $incompatible_core_version_requirements[] = $row->core_version_requirement;
      }
    }
    catch (Exception $e) {
      drush_log(dt('Exception for "@core_version_requirement": @message', [
        '@core_version_requirement' => $row->core_version_requirement,
        '@message' => $e->getMessage(),
      ]));
    }
  }
  drush_log(dt('Core version requirements that do not match @core_constraint:', [
    '@core_constraint' => $core_constraint,
  ]));
  drush_log('"' . implode('", "', $incompatible_core_version_requirements) . '"');

  $result = db_query("SELECT n.nid, n.title, prsv.branch, pcc.core_version_requirement FROM {project_release_supported_versions} prsv INNER JOIN {node} n ON n.nid = prsv.nid AND n.status = 1 AND n.type != 'project_core' INNER JOIN {field_data_field_release_category} fdf_rc ON fdf_rc.entity_id = prsv.latest_release AND fdf_rc.field_release_category_value = 'current' INNER JOIN {project_composer_component} pcc ON pcc.release_nid = fdf_rc.entity_id AND pcc.component_role = 'primary' WHERE prsv.supported = 1 AND pcc.core_version_requirement IN (:incompatible_core_version_requirements) GROUP BY prsv.nid, prsv.branch", [
    ':incompatible_core_version_requirements' => $incompatible_core_version_requirements,
  ])->fetchAll(PDO::FETCH_ASSOC);
  $count = count($result);
  drush_log(PHP_EOL . dt('@count branches to mark unsupported:', ['@count' => $count]));
  drush_print(drush_format_table($result));

  if (drush_get_option('maintainers-for-email')) {
    $filename = '/tmp/unsupport-maintainers-' . date('c') . '.csv';
    drush_log(dt('Writing @filename', ['@filename' => $filename]));
    db_query('SET SESSION group_concat_max_len = 65535');
    $result = db_query("SELECT u.name, u.mail, group_concat(DISTINCT concat(n.title, ' ', prsv.branch, '*') SEPARATOR ', ') project_branches, group_concat(DISTINCT concat('https://www.drupal.org/', ua.alias) SEPARATOR ', ') project_urls FROM project_release_supported_versions prsv INNER JOIN node n ON n.nid = prsv.nid AND n.status = 1 AND n.type != 'project_core' INNER JOIN field_data_field_release_category fdf_rc ON fdf_rc.entity_id = prsv.latest_release AND fdf_rc.field_release_category_value = 'current' INNER JOIN project_composer_component pcc ON pcc.release_nid = fdf_rc.entity_id AND pcc.component_role = 'primary' INNER JOIN url_alias ua ON ua.source = concat('node/', n.nid) INNER JOIN project_release_project_maintainer prpm ON prpm.nid = n.nid AND prpm.administer_releases = 1 INNER JOIN users u ON u.uid = prpm.uid AND u.status = 1 WHERE supported = 1 AND pcc.core_version_requirement IN (:incompatible_core_version_requirements) GROUP BY u.uid", [
      ':incompatible_core_version_requirements' => $incompatible_core_version_requirements,
    ], ['fetch' => PDO::FETCH_ASSOC]);
    $stream = fopen($filename, 'w');
    fputcsv($stream, ['username', 'email', 'project branches', 'project urls']);
    foreach ($result as $row) {
      fputcsv($stream, $row);
    }
    fclose($stream);
  }

  if (drush_get_option('no-dry-run') && drush_confirm(dt('@count branches WILL BE marked unsupported', ['@count' => $count]))) {
    drush_log(dt('Marking releases unsupported…'));
    db_query("UPDATE {project_release_supported_versions} prsv INNER JOIN {node} n ON n.nid = prsv.nid AND n.status = 1 AND n.type != 'project_core' INNER JOIN {field_data_field_release_category} fdf_rc ON fdf_rc.entity_id = prsv.latest_release AND fdf_rc.field_release_category_value = 'current' INNER JOIN {project_composer_component} pcc ON pcc.release_nid = fdf_rc.entity_id AND pcc.component_role = 'primary' SET prsv.supported = 0, prsv.recommended = 0 WHERE prsv.supported = 1 AND pcc.core_version_requirement IN (:incompatible_core_version_requirements)", [
    ':incompatible_core_version_requirements' => $incompatible_core_version_requirements,
  ]);
    cache_clear_all('*', 'cache_project_release_download_table', TRUE);
  }
}

/**
 * Remove users from keycloak that no longer exist.
 */
function drush_drupalorg_sync_keycloak_user_delete() {
  $query = db_select('authmap', 'map')
    ->fields('map', ['uid']);
  $query->leftJoin('users', 'u', 'u.uid=map.uid AND map.module="openid_connect_keycloak"');
  $query->isNull('u.uid');
  foreach ($query->execute()->fetchCol() as $uid) {
    (new KeycloakIntegration())->onUserDelete($uid);
  }
}

/**
 * Get keycloak user info by its UUID.
 */
function drush_drupalorg_keycloak_user_by_uuid($uuid) {
  echo (new KeycloakIntegration())->getUser($uuid) . PHP_EOL;
}

function drush_drupalorg_keycloak_admin_users_integrity() {
  $query = db_select('users', 'u')
    ->condition('r.rid', drupalorg_2fa_required_roles())
    ->fields('u', ['uid']);
  $query->leftJoin('users_roles', 'r', 'r.uid=u.uid');

  $uids = $query->execute()->fetchCol();
  printf("Starting integrity check of (%d) users.\n", count($uids));
  $integration = new KeycloakIntegration();
  foreach ($uids as $uid) {
    $integration->integrityCheckUser($uid);
  }
  printf("Finished integrity check of (%d) users.\n", count($uids));
}

function drush_drupalorg_keycloak_random_users_integrity() {
  $query = db_query('SELECT uid FROM (SELECT @row := @row +1 AS rownum, uid FROM ( SELECT @row :=0) r, users ) ranked WHERE rownum % 400 = 2');
  $integration = new KeycloakIntegration();

  $uids = $query->fetchCol();
  printf("Starting integrity check of (%d) users.\n", count($uids));
  foreach ($uids as $uid) {
    $integration->integrityCheckUser($uid);
  }
  printf("Finished integrity check of (%d) users.\n", count($uids));
}

function drush_drupalorg_keycloak_users_count() {
  printf("Starting user counting.\n");
  $drupal_users_count = db_select('users')
    ->countQuery()
    ->execute()
    ->fetchField();
  $drupal_authmap_count = db_select('authmap')
    ->countQuery()
    ->execute()
    ->fetchField();
  $integration = new KeycloakIntegration();
  $keycloak_users_count = $integration->countUsers();
  if ($drupal_users_count == $keycloak_users_count && $drupal_users_count == $drupal_authmap_count) {
    printf("Total user count matches: %d.\n", $drupal_users_count);
  }
  else {
    printf("Total user counts does not match.\n");
    printf("Drupal users: %d.\n", $drupal_users_count);
    printf("Drupal authmap: %d.\n", $drupal_authmap_count);
    printf("Keycloak users: %d.\n", $keycloak_users_count);
  }
}
